from __future__ import division

from pyglet.window import key
from pyglet import graphics
from pyglet.gl import *
import data

import random # very carefully

from vector import Vector as v
import sound

from constants import *

class TimeShiftObject(object):
    rate = 1.0
    ticks = 0
    subticks = 0

    def tick(self):
        self.ticks += 1
        self.realtick()

    def realtick(self):
        pass

class TimeShiftEntity(TimeShiftObject):
    old_pos = v((100, 100))
    pos = v((100, 100))

    dead = False
    echo = False
    sprite = None
    img = None
    
    friction = 0.8

    def __init__(self, handler):
        self.h = handler

    def realtick(self):
        o = self.pos
        self.pos += (self.pos - self.old_pos) * self.friction
        self.move()
        self.old_pos = o

    def move(self):
        pass

    def teleport(self, pos):
        self.pos = pos
        self.old_pos = pos

class SolidEntity(TimeShiftEntity):
    bumped = False
    within = False
    heading = 0.0
    radius = 5
    scale = 1.0
    img = "box.png"

    def __init__(self, *args, **kwds):
        super(SolidEntity, self).__init__(*args, **kwds)
        self.tex = data.load_texture(self.img)
        self.vlist = None

    def move(self):
        o2 = self.pos

        self.pos = self.h.level.push(self.pos, self.radius)
        if self.within:
            self.pos = self.h.level.push_in(self.pos, self.radius)
        else:
            if self.h.level.contains(self.pos, self.radius):
                self.within = True

        self.bumped = (self.pos != o2)

    def realtick(self):
        super(SolidEntity, self).realtick()
        if not self.dead:
            self.update_vlist()

    def get_color(self):
        if self.echo:
            return [255,255,255,127]
        return [255] * 4

    def create_vlist(self):
        import gamemode
        tc = self.tex.tex_coords
        vdata = [0] * 8
        cdata = self.get_color() * 4
        tdata = [tc[0], tc[1], tc[3], tc[4], tc[6], tc[7], tc[9], tc[10]]
        group = gamemode.OrderTextureGroup(4, self.tex)
        self.vlist = self.h.batch.add(4, GL_QUADS, group, ('v2f', vdata), ('t2f', tdata), ('c4B', cdata))

    def update_vlist(self):
        if self.vlist is None:
            self.create_vlist()

        from math import sin, cos, radians
        x, y = self.pos
        h = radians(self.heading)
        r = self.radius
        rc = r * self.scale * cos(h)
        rs = r * self.scale * sin(h)
        self.vlist.vertices[:] = [
                x-rc+rs, y-rs-rc,
                x+rc+rs, y+rs-rc,
                x+rc-rs, y+rs+rc,
                x-rc-rs, y-rs+rc,
        ]

    def destroy(self):
        if self.vlist:
            self.vlist.delete()

class Enemy(SolidEntity):
    heading = 0
    hp = 3
    radius = 5
    dps = 0.1
    vel = 0.3
    points = 10
    colour = (255, 255, 255)
    kill_angle = 0
    def __init__(self, handler, randomseed):
        super(Enemy, self).__init__(handler)
        self.rng = random.Random(handler.ticks + randomseed)
    def start(self):
        self.heading = (v((320, 240)) - self.pos).angle
        self.update_vlist()
    def can_attack(self):
        return self.h.level.in_attack_zone(self.pos)
    def damage(self, amt, should_score, heading, silent=False):
        if self.dead: return
        self.hp -= amt
        if self.hp <= 0:
            self.dead = True
            if not silent:
                sound.destroy()
            if should_score:
                self.h.score_points(self.points)
            if self.vlist:
                self.vlist.delete()
                self.vlist = None
            for ii in xrange(5):
                d = v((self.radius, 0)).rotated(random.randrange(360))
                self.h.add_fragment(self.rate, self.pos + d, heading)


    def get_color(self):
        if self.echo:
            return self.colour + (127,)
        return self.colour + (255,)

class Bumbler(Enemy):
    img = "red.png"
    def move(self):
        dir = v((self.vel, 0)).rotated(self.heading)
        self.pos += dir
        if self.bumped: 
            self.heading = self.rng.randrange(360)
        super(Enemy, self).move()
            
class BumblingSeeker(Bumbler):
    points = 15
    img = "blue.png"
    def move(self):
        super(BumblingSeeker, self).move()
        if self.bumped: 
            self.heading = self.h.level.get_path_dir(self.pos).angle

class Robot(Bumbler):
    img = "green.png"
    def move(self):
        super(Robot, self).move()
        if self.ticks % 50 == 0 and self.rng.randrange(3) == 0:
            self.heading = self.h.level.get_path_dir(self.pos).angle
        if self.within:
            self.heading = 90 * ((self.heading + 45) // 90)

class Darter(Enemy):
    points = 25
    period = 45
    img = "pink.png"
    @property
    def vel(self):
        return 0.45 - (self.ticks % self.period) / 100
    def move(self):
        super(Darter, self).move()
        dir = v((self.vel, 0)).rotated(self.heading)
        self.pos += dir
        if self.ticks % self.period == 0 and self.within:
            if self.rng.randrange(2) == 1:
                self.heading = self.h.level.get_path_dir(self.pos).angle
            else:
                self.heading = self.rng.randrange(360)
    def start(self):
        self.period = self.rng.randrange(40, 45)
        super(Darter, self).start()


class Whirly(BumblingSeeker):
    img = "purple.png"
    def move(self):
        super(Whirly, self).move()
        if self.within:
            self.heading += 1

class Seeker(Enemy):
    points = 30
    img = "cyan.png"
    def move(self):
        dir = v((self.vel, 0)).rotated(self.heading)
        self.pos += dir
        if (self.bumped or self.ticks % 20 == 0) and self.within:
            self.heading = self.h.level.get_path_dir(self.pos).angle
        super(Seeker, self).move()
        
class Sniper(Enemy):
    points = 75
    vel = 0.15
    old_d = 1e8
    dps = 0.06
    img = "lilac.png"
    def move(self):
        dir = v((self.vel, 0)).rotated(self.heading)
        self.pos += dir
        if self.within and self.ticks % 10 == 0:
            c = self.h.level.core
            d = (self.pos - v((c.xm, c.ym))).length2
            if self.can_attack() or d > self.old_d - 100:
                self.heading = self.rng.randrange(360)
            else:
                self.heading = self.h.level.get_path_dir(self.pos).angle
            self.heading = 90 * ((self.heading - 22.5) // 90) + 45
            self.old_d = d
        super(Sniper, self).move()
    def can_attack(self):
        return self.h.level.in_attack_zone(self.pos, a_range=108)

class DriftSeeker(Bumbler):
    points = 30
    img = "cyan.png"
    def start(self):
        super(DriftSeeker, self).start()
        self.heading = self.rng.randrange(360)
        self.update_vlist()
    def realtick(self):
        super(DriftSeeker, self).realtick()
        if self.ticks > 20:
            self.h.new_enemies.append((Seeker, self.pos, 0, self.echo))
            self.dead = True
            if self.vlist:
                self.vlist.delete()
                self.vlist = None

class Breeder(Bumbler):
    points = 200
    img = "yellow.png"
    def realtick(self):
        super(Breeder, self).realtick()
        if self.ticks % 150 == 50:
            self.h.new_enemies.append((Seeker, self.pos, self.ticks, self.echo))

class Birther(Bumbler):
    points = 750
    vel = 0.15
    img = "orange.png"
    def realtick(self):
        super(Birther, self).realtick()
        if self.ticks > 201 and self.ticks % 200 > 188:
            self.h.new_enemies.append((DriftSeeker, self.pos, self.ticks, self.echo))
                
class PlayerShip(SolidEntity):
    heading = 0
    firing = False
    firetimer = 0
    scale = 1.2
    img = "ship.png"
    def __init__(self, handler, record=None):
        super(PlayerShip, self).__init__(handler)
        if record is None:
            self.recording = True
            self.rec = ControlRecorder()
        else:
            self.recording = False
            self.rec = record
            record.rewind()
        self.history = []

    def start(self):
        self.update_vlist()

    def move(self):
        accel, _1, _2 = self.update_controls()
        self.pos += accel
        if not self.firing:
            delta = (self.pos - self.old_pos)
            if delta.length2 > 0.1:
                self.heading = delta.angle
        super(PlayerShip, self).move()
            
    def realtick(self):
        super(PlayerShip, self).realtick()
        if self.firing and self.firetimer <= 0:
            self.h.add_projectile(self.rate, self.pos, self.heading - 2, self.echo)
            self.h.add_projectile(self.rate, self.pos, self.heading + 2, self.echo)
            self.firetimer = FIRING_RATE
            did_fire = True
        else:
            self.firetimer -= 1
            did_fire = False
        if DEBUG_POSLOG:
            self.history.append((self.pos, self.heading, did_fire, self.bumped))
        
    def update_controls(self):
        if self.recording:
            accel = v((0, 0))
            if self.k[key.UP] or self.k[key.W]:    accel += v((0, 1))
            if self.k[key.DOWN] or self.k[key.S]:  accel += v((0, -1))
            if self.k[key.LEFT] or self.k[key.A]:  accel += v((-1, 0))
            if self.k[key.RIGHT] or self.k[key.D]: accel += v((1, 0))
            state = (accel, self.k.fire, self.k[key.LSHIFT])
            self.rec.save_tick(self.ticks, state)
            self.firing = state[1]
            return state
        else:
            res = self.rec.get_tick(self.ticks)
            self.firing = res[1]
            return res
        
class ControlRecorder(object):
    def __init__(self):
        self.last_state = (v((0, 0)), False, False)
        self.history = {}
        
    def rewind(self):
        self.last_state = (v((0, 0)), False, False)

    def stop(self, tick):
        self.history[tick] = (v((0, 0)), False, False)
        
    def save_tick(self, tick, state):
        if state == self.last_state:
            pass
        else:
            self.history[tick] = state
            self.last_state = state
            
    def get_tick(self, tick):
        if tick in self.history:
            self.last_state = self.history[tick]
        return self.last_state

class Projectile(TimeShiftEntity):
    s = 1.8
    damage = 1
    color = (1, 1, 0, 1)
    def __init__(self, handler, heading):
        super(Projectile, self).__init__(handler)
        self.accel = v((self.s, 0)).rotated(heading)
        self.heading = heading
        
    def move(self):
        self.pos += self.accel
        self.old_pos = self.pos
        if self.h.level.solid(self.pos):
            self.dead = True
        
class Fragment(Projectile):
    s = 4
    damage = 0
    friction = 0.95
    color = (0.5, 0.5, 0.5, 1)
    def __init__(self, handler, heading):
        self.s = random.uniform(2, 3)
        super(Fragment, self).__init__(handler, heading + random.randrange(-10, 11))
        self.life = random.randrange(15, 25)
    def move(self):
        super(Fragment, self).move()
        self.accel = v((0, 0))
        self.life -= 1
        if self.life < 1:
            self.dead = True
