"""Attribute descriptors and caching properties.

The behaviour of attribute descriptors and caching properties defined here is
quite magical and a brief discussion may be useful in comprehending it.

The objective is to create a system whereby any of the data used to configure a
panel can be changed and the panel is still able to cache lengthy or complex
computations based on those data. The solution therefore is a mechanism to
track which of those data are used during which cached computations, and a
further mechanism to remove the necessary cached values whenever those data
are changed.

These goals are accomplished by the interactions of the `Attribute` and `Cached`
classes. The `Attribute` class is a descriptor which is used to define the
significant data that will be used in cached computations. The `Cached` class
is a decorator and a descriptor which is used to wrap any method that represents
a cached computation.

When an instance of `Cached` is accessed the cached value is first looked for
in a dictionary stored on the accessing `Panel` object. If it is not found the
underlying method is called and the fact that the value is being computed is
recorded globally. Thereafter should an instance of `Attribute` be accessed
the attribute will record the fact that the instance of `Cached` accessed on
the accessing object is dependent on its value.

What follows is a brief example to show the desirable aspects of the behaviour:

import interface

class Simple(interface.panel.Panel):

    myattr = interface.attribute.Attribute(0)
    myotherattr = interface.attribute.Attribute("Hello, World!")

    @interface.attribute.Cached
    def mycached(self):
        print "mycached computed"
        return 2 ** self.myattr

simple = Simple()
assert simple.mycached == 1 # prints "mycached computed"
assert simple.mycached == 1 # prints nothing
simple.myattr += 2
assert simple.mycached == 4 # prints "mycached computed"
assert simple.mycached == 4 # prints nothing
simple.myotherattr = None
assert simple.mycached == 4 # prints nothing

"""

import weakref

from interface import utils

__all__ = [
    "Attribute",
    "Directional",
    "Tracking",
    "Cached",
]


###############################################################################
## CACHING PROPERTIES & ATTRIBUTE DESCRIPTORS                                ##
###############################################################################

class _DependentTracker(object):
    """Mixin for objects tracking dependent caching properties on panels.

    """

    _computing = []

    def _register_dependent(self, obj):
        """Register the top caching property computation as a dependent.

        :Parameters:
            `obj` : object
                The object which should be depended on.

        """
        if self._computing:
            top = self._computing[-1]
            if self not in obj._dependents:
                obj._dependents[self] = set()
            obj._dependents[self].add(top)

    def _clear_dependents(self, obj):
        """Clear and remove all dependent caching properties.

        :Parameters:
            `obj` : object
                The object whose dependents should be cleared.

        """
        if self in obj._dependents:
            dependents = obj._dependents[self]
            while len(dependents) > 0:
                prop, dobj = dependents.pop()
                prop.__delete__(dobj())

    def __repr__(self):
        return "<%s : %s>" % (utils.get_cls_name(self), self._name)


class Attribute(_DependentTracker):
    """Panel attribute descriptors.

    """

    def __init__(self, default, dynamic=False, dtype=None):
        """Create an Attribute object.

        :Parameters:
            `default` : object
                The default value of the attribute.
            `dynamic` : boolean
                If set then dependents will not be managed.
            `dtype` : type
                If set the stored value will be coerced.

        """
        super(Attribute, self).__init__()
        self._dtype = dtype or (lambda v: v)
        self._default = self._process(default)
        self._name = None

        if dynamic:
            self._register_dependent = lambda obj: None
            self._clear_dependents = lambda obj: None

    def __get__(self, obj, cls):
        """Retrieve the value from the panel attribute dictionary.

        """
        if obj is None:
            return self
        self._register_dependent(obj)
        attrs = obj._panel_attrs
        try:
            return attrs[self._name]
        except KeyError:
            value = self._default
            attrs[self._name] = value
            return self._default

    def __set__(self, obj, value):
        """Set the value in the panel attribute dictionary removing dependent
        caching properties if necessary.

        """
        value = self._process(value)
        attrs = obj._panel_attrs
        if attrs.get(self._name, self._default) != value:
            attrs[self._name] = value
            self._clear_dependents(obj)

    def __delete__(self, obj):
        """Reset the value in the panel attribute dictionary to the default
        removing dependent caching properties if necessary.

        """
        self.__set__(obj, self._default)

    def _process(self, value):
        """Process the value into the stored form.

        :Parameters:
            `value` : object
                The object being set.

        """
        if value is None:
            return value
        return self._dtype(value)


class Directional(Attribute):
    """Attribute accepting values in four directions.

    """

    def _process(self, value):
        """Process the value into the stored form.

        :Parameters:
            `value` : object
                The object being set.

        """
        if value is None:
            return value
        try:
            value = map(self._dtype, value)
        except TypeError:
            return (self._dtype(value),) * 4
        if len(value) == 2:
            return tuple(value * 2)
        elif len(value) == 4:
            return tuple(value)
        raise ValueError


class Tracking(Attribute):
    """Attribute which follows the value of another attribute.

    """

    def __init__(self, tracked, **kwds):
        """Create a Tracking object.

        :Parameters:
            `tracked` : str
                The name of the attribute to track.

        """
        super(Tracking, self).__init__(None, **kwds)
        self._tracked = tracked

    def __get__(self, obj, cls):
        """Retrieve the value from the panel attribute dictionary.

        """
        if obj is None:
            return self
        self._register_dependent(obj)
        attrs = obj._panel_attrs
        if attrs.get(self._name) is not None:
            return attrs[self._name]
        return getattr(obj, self._tracked)


class Ratio(Attribute):
    """Attribute accepting values as a ratio of two numbers.

    """

    def _process(self, value):
        """Process the value into the stored form.

        :Parameters:
            `value` : object
                The object being set.

        """
        if value is None:
            return value
        try:
            value = map(self._dtype, value)
        except ValueError:
            value = map(self._dtype, (value, 1.0))
        if len(value) == 2:
            return tuple(value)
        raise ValueError

class Cached(_DependentTracker):
    """Caching property objects.

    """

    def __init__(self, function):
        """Create a Cached object.

        :Parameters:
            `function` : function
                The wrapped method.

        """
        super(Cached, self).__init__()
        self._name = function.func_name
        self._function = function

    def __get__(self, obj, cls=None):
        """Optionally compute and then return the cached value.

        """
        if obj is None:
            return self
        self._register_dependent(obj)
        cache = obj._panel_cache
        try:
            return cache[self._name]
        except KeyError:
            dobj = weakref.ref(obj)
            Cached._computing.append((self, dobj))
            try:
                cache[self._name] = self._function(obj)
            finally:
                Cached._computing.pop()
            return cache[self._name]

    def __set__(self, obj, value):
        """Override the cached value.

        """
        if self._name not in obj._panel_cache or \
            obj._panel_cache[self._name] != value:
                obj._panel_cache[self._name] = value
                self._clear_dependents(obj)

    def __delete__(self, obj):
        """Remove the cached value, forcing recomputation.

        """
        if obj is None:
            return
        try:
            del obj._panel_cache[self._name]
            self._clear_dependents(obj)
        except KeyError:
            pass
