"""Basic content panel definitions.

"""

import pyglet
from pyglet.gl import *

from interface import attribute
from interface import panel
from interface import utils

__all__ = [
    "Root",
    "Color",
    "Text",
    "Image",
]


###############################################################################
## CONTENT PANEL IMPLEMENTATIONS                                             ##
###############################################################################

class Root(panel.Panel):
    """Panel for attaching a window.

    """

    window = attribute.Attribute(None)

    @attribute.Cached
    def parent_size(self):
        if self.window is None:
            return (0.0, 0.0)
        return self.window.get_size()

    @attribute.Cached
    def layout_size(self):
        if self.window is None:
            return (0.0, 0.0)
        return self.window.get_size()

    def draw(self):
        if self.layout_size != self.window.get_size():
            self._clear_cached_attributes()
        super(Root, self).draw()


class Color(panel.Panel):
    """Panel containing a block of color.

    """

    color = attribute.Attribute((0.4, 0.4, 0.8, 0.5))

    def draw_content(self):
        width, height = self.content_size
        cattr = 'c%dB' % len(self.color)
        cdata = [int(255 * x) for x in self.color] * 4
        vdata = [0.0, 0.0, width, 0.0, width, height, 0.0, height]
        pyglet.graphics.draw(4, GL_QUADS, ('v2f', vdata), (cattr, cdata))


class Text(panel.Panel):
    """Panel containing some text.

    """

    font_name = attribute.Attribute(None)
    font_size = attribute.Attribute(0.08) # rel: root.height
    color = attribute.Attribute((1.0, 1.0, 1.0, 1.0))
    italic = attribute.Attribute(False)
    bold = attribute.Attribute(False)
    align = attribute.Attribute("center")
    text = attribute.Attribute(u"")

    def init_panel(self):
        self.document = d = pyglet.text.document.UnformattedDocument()
        self.layout = pyglet.text.layout.IncrementalTextLayout(d, 0, 0)

    def compute_content_size(self):
        self.layout.begin_update()
        self.layout.content_width = 0 # don't worry about this
        rcwidth, rcheight = self.root.content_size
        mvwidth, mvheight = self.max_view_size
        self.document.set_style(0, 0, {
            "font_name" : self.font_name,
            "font_size" : self.font_size * rcheight,
            "color" : [int(255 * x) for x in self.color],
            "italic" : self.italic,
            "bold" : self.bold,
            "align" : self.align})
        self.document.text = self.text
        self.layout.x = self.layout.y = 0
        self.layout.width = mvwidth
        self.layout.height = mvheight
        self.layout.content_valign = "bottom"
        self.layout.multiline = True
        self.layout.end_update()
        return self.layout.content_width, \
               self.layout.content_height

    def draw_content(self):
        glPushMatrix()
        blank_width = self.layout.width - self.layout.content_width
        offset = -self._symbol_factors[self.align] * blank_width
        glTranslatef(offset, 0.0, 0.0)
        self.layout.draw()
        glPopMatrix()


class Image(panel.Panel):
    """Panel containing an image.

    """

    image = attribute.Attribute(None)
    constrain = attribute.Attribute(True, dtype=bool)

    def compute_content_size(self):
        width, height = self.max_view_size
        if self.constrain:
            iwidth = self.image.width
            iheight = self.image.height
            iaspect = iwidth / iheight
            if width < height * iaspect:
                height = width / iaspect
            elif width >= height * iaspect:
                width = height * iaspect
        return width, height

    def draw_content(self):
        cwidth, cheight = self.content_size
        scale_x = cwidth / self.image.width
        scale_y = cheight / self.image.height
        glPushMatrix()
        glScalef(scale_x, scale_y, 1.0)
        self.image.blit(0, 0)
        glPopMatrix()
