"""Panel signal/event framework.

"""

import pyglet

from interface import METAPOD_DEBUG
from interface import utils

__all__ = [
    "Handler",
]


###############################################################################
## DEFAULT PANEL EVENT HANDLER                                               ##
###############################################################################

class Handler(object):
    """Panel event-signal conversion object.

    """

    ## Constructor
    ##############

    def __init__(self, root):
        """Create a Handler object.

        :Parameters:
            `root` : interface.panel.Panel
                The panel at the root of the panel tree.

        """
        self.root = root

    ## Event handlers
    #################

    @staticmethod
    def report_event():

        # Frame introspection.
        event_frame = sys._getframe().f_back
        event_name = event_frame.f_code.co_name
        event_varnames = event_frame.f_code.co_varnames
        event_vars = event_frame.f_locals

        # Alternative 'repr' functions.
        from pyglet.window import key, mouse
        reprs = { "modifiers" : key.modifiers_string,
                  "motion"    : key.motion_string,
                  "symbol"    : key.symbol_string,
                  "button"    : mouse.buttons_string,
                  "buttons"   : mouse.buttons_string, }

        # Construct argument lists.
        arguments = []
        for name in event_varnames[1:]:
            value = reprs.get(name, repr)(event_vars[name])
            arguments.append("%s=%s" % (name, value))

        # Print event.
        print "%s(%s)" % (event_name, ", ".join(arguments))

    def on_activate(self):
        """Process the 'on_activate' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_close(self):
        """Process the 'on_close' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_context_lost(self):
        """Process the 'on_context_list' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_context_state_lost(self):
        """Process the 'on_context_state_lost' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_deactivate(self):
        """Process the 'on_deactivate' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_draw(self):
        """Process the 'on_draw' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_expose(self):
        """Process the 'on_expose' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_hide(self):
        """Process the 'on_hide' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_key_press(self, symbol, modifiers):
        """Process the 'on_key_press' event.

        """
        if self.root.process_signal("key_press", symbol, modifiers):
            return pyglet.event.EVENT_HANDLED
        return pyglet.event.EVENT_UNHANDLED

    def on_key_release(self, symbol, modifiers):
        """Process the 'on_key_release' event.

        """
        if self.root.process_signal("key_release", symbol, modifiers):
            return pyglet.event.EVENT_HANDLED
        return pyglet.event.EVENT_UNHANDLED

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        """Process the 'on_mouse_drag' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_mouse_enter(self, x, y):
        """Process the 'on_mouse_enter' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_mouse_leave(self, x, y):
        """Process the 'on_mouse_leave' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_mouse_motion(self, x, y, dx, dy):
        """Process the 'on_mouse_motion' event.

        """
        if METAPOD_DEBUG:
            self.root.process_contained_signal("debug_focus", x, y)
        self.root.process_mouse_signal("mouse_move", x, y)
        return pyglet.event.EVENT_UNHANDLED

    def on_mouse_press(self, x, y, button, modifiers):
        """Process the 'on_mouse_press' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_mouse_release(self, x, y, button, modifiers):
        """Process the 'on_mouse_release' event.

        """
        if self.root.process_contained_signal("mouse_click", x, y, modifiers):
            return pyglet.event.EVENT_HANDLED
        return pyglet.event.EVENT_UNHANDLED

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        """Process the 'on_mouse_scroll' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_move(self, x, y):
        """Process the 'on_move' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_resize(self, width, height):
        """Process the 'on_resize' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_show(self):
        """Process the 'on_show' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_text(self, text):
        """Process the 'on_text' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_text_motion(self, motion):
        """Process the 'on_text_motion' event.

        """
        return pyglet.event.EVENT_UNHANDLED

    def on_text_motion_select(self, motion):
        """Process the 'on_text_motion_select' event.

        """
        return pyglet.event.EVENT_UNHANDLED
