"""Basic layout panel definitions.

"""

from interface import attribute
from interface import panel
from interface import utils

__all__ = [
    "Horizontal",
    "Vertical",
    "Grid",
]


###############################################################################
## LAYOUT PANEL IMPLEMENTATIONS                                              ##
###############################################################################

class Horizontal(panel.Layout):
    """Layout children horizontally.

    """

    spacing = attribute.Attribute(0.0) # rel: layout.height

    def compute_content_size(self):
        lwidth, lheight = self.layout_size
        spacing = self.spacing * lheight
        current_width = max_height = 0.0
        for child in self.children:
            child.halign = "left"
            child.anchor_x = None
            child.align_xoffset = current_width
            cwidth, cheight = child.content_size
            current_width += cwidth + spacing
            max_height = max(max_height, cheight)
        return current_width - spacing, max_height


class Vertical(panel.Layout):
    """Layout children vertically.

    """

    spacing = attribute.Attribute(0.0) # rel: layout.height

    def compute_content_size(self):
        lwidth, lheight = self.layout_size
        spacing = self.spacing * lheight
        current_height = max_width = 0.0
        for child in reversed(self.children):
            child.valign = "bottom"
            child.anchor_y = None
            child.align_yoffset = current_height
            cwidth, cheight = child.content_size
            current_height += cheight + spacing
            max_width = max(max_width, cwidth)
        return max_width, current_height - spacing


class Grid(panel.Layout):
    """Layout children in a fixed width grid.

    """

    spacing = attribute.Attribute(0.0) # rel: layout.height
    grid_width = attribute.Attribute(5, dtype=int)

    def compute_content_size(self):
        rows = [0.0] * -(len(self.children) // -self.grid_width)
        cols = [0.0] * self.grid_width
        for idx, child in enumerate(self.children):
            child.anchor_x = child.anchor_y = None
            cwidth, cheight = child.content_size
            row, col = divmod(idx, self.grid_width)
            row = len(rows) - row - 1
            rows[row] = max(cheight, rows[row])
            cols[col] = max(cwidth, cols[col])
        lwidth, lheight = self.layout_size
        spacing = self.spacing * lheight
        for idx, child in enumerate(self.children):
            row, col = divmod(idx, self.grid_width)
            row = len(rows) - row - 1
            child.align_width = cols[col]
            child.align_height = rows[row]
            child.align_xoffset = sum(cols[:col]) + spacing * col
            child.align_yoffset = sum(rows[:row]) + spacing * row
        return sum(cols) + spacing * (len(cols) - 1), \
               sum(rows) + spacing * (len(rows) - 1)
