"""Base panel definitions.

"""

from __future__ import division

import inspect
import math
import textwrap

import pyglet
from pyglet.gl import *

from interface import METAPOD_DEBUG
from interface import attribute
from interface import utils

__all__ = [
    "Panel",
    "Layout",
]


###############################################################################
## DEBUG FUNCTIONS                                                           ##
###############################################################################

if METAPOD_DEBUG:

    def _pointtest(x, y, c=(1,0,0,1)):
        glPushAttrib(GL_ALL_ATTRIB_BITS)
        glPointSize(5.0)
        glColor4f(*c)
        glBegin(GL_POINTS)
        glVertex2f(x, y)
        glEnd()
        glPopAttrib()

    def _recttest(x, y, w, h, c=(1,0,0,1)):
        glPushAttrib(GL_ALL_ATTRIB_BITS)
        glColor4f(*c)
        glBegin(GL_LINE_LOOP)
        glVertex2f(x, y)
        glVertex2f(x + w, y)
        glVertex2f(x + w, y + h)
        glVertex2f(x, y + h)
        glEnd()
        glPopAttrib()


###############################################################################
## PANEL BASE CLASS (AND METACLASS)                                          ##
###############################################################################

class Panel(object):
    """Base class for interface objects.

    """

    ## Keyword attributes
    #####################

    # Dynamic offset.
    xoffset = attribute.Attribute(0, dynamic=True, dtype=int)
    yoffset = attribute.Attribute(0, dynamic=True, dtype=int)

    # Alignment offset.
    align_xoffset = attribute.Attribute(0, dynamic=True, dtype=int)
    align_yoffset = attribute.Attribute(0, dynamic=True, dtype=int)

    # Scroll offset.
    scroll_xoffset = attribute.Attribute(0, dynamic=True, dtype=int)
    scroll_yoffset = attribute.Attribute(0, dynamic=True, dtype=int)

    # Dimensions.
    width = attribute.Attribute(None) # rel: layout.width
    height = attribute.Attribute(None) # rel: layout.height
    aspect = attribute.Ratio(None)

    # Alignment dimensions.
    align_width = attribute.Attribute(None) # rel: absolute
    align_height = attribute.Attribute(None) # rel: absolute

    # Alignment.
    halign = attribute.Attribute("center") # rel: parent.width
    valign = attribute.Attribute("center") # rel: parent.height
    anchor_x = attribute.Tracking("halign") # rel: content.width
    anchor_y = attribute.Tracking("valign") # rel: content.height

    # Directional padding.
    margin = attribute.Directional(0.0) # rel: layout.height
    padding = attribute.Directional(0.0) # rel: layout.height

    # Frame.
    frame_width = attribute.Attribute(0.02) # rel: layout.height
    frame_texture = attribute.Attribute(None)
    frame_color = attribute.Attribute(None)

    # Background.
    background_texture = attribute.Attribute(None)
    background_color = attribute.Attribute(None)

    # Scissor test.
    scissor = attribute.Attribute(False)

    ## Metaclass
    ############

    class PanelMeta(type):
        """Metaclass for interface objects.

        """

        def __init__(self, name, bases, dict):
            """Create a PanelMeta object.

            """
            super(self.PanelMeta, self).__init__(name, bases, dict)

            # Start from blank name trees.
            self._keyword_names = self._cached_names = frozenset()

            # Collate the name trees from base classes.
            for base in bases:
                self._keyword_names |= getattr(base, "_keyword_names", set())
                self._cached_names |= getattr(base, "_cached_names", set())

            # Process attribute objects and cached objects.
            for name, value in dict.items():
                if isinstance(value, attribute.Attribute):
                    self._keyword_names |= frozenset([name])
                    value._name = name
                if isinstance(value, attribute.Cached):
                    self._cached_names |= frozenset([name])

            # Check all names against the class.
            for name in set(self._keyword_names):
                if not isinstance(getattr(self, name), attribute.Attribute):
                    self._keyword_names -= frozenset([name])
            for name in set(self._cached_names):
                if not isinstance(getattr(self, name), attribute.Cached):
                    self._cached_names -= frozenset([name])

    __metaclass__ = PanelMeta

    ## Constructor
    ##############

    def __init__(self, **kwds):
        """Create a Panel object.

        """
        super(Panel, self).__init__()

        #: Root of this panel tree.
        self._root = self
        #: Parent of this panel.
        self._parent = None
        #: List of children of this panel.
        self._children = []
        #: Whether to render this panel.
        self._visible = True

        # Create a blank panel state.
        self._panel_attrs = {}
        self._panel_cache = {}
        self._dependents = {}

        # Call the subclass init hook.
        self.init_panel()

        # Populate the panel state.
        self._load_keyword_attributes(**kwds)

    def _load_keyword_attributes(self, **kwds):
        self._visible = kwds.pop("visible", True)
        for name, value in kwds.items():
            if name not in self._keyword_names:
                raise TypeError, "%s got an unexpected keyword argument %r" % \
                        (utils.get_cls_name(self), name)
            setattr(self, name, value)

    def _clear_cached_attributes(self):
        for node in self.descendants:
            for name in self._cached_names:
                delattr(self, name)

    def __repr__(self):
        return "<%s : %s>" % (utils.get_cls_name(self), hex(id(self)))

    ## Extension API
    ################

    def init_panel(self):
        """Prepare panel resources.

        """

    def compute_content_size(self):
        """Compute the size of the panel content.

        """

    def draw_content(self):
        """Render the panel content.

        """

    ## Properties
    #############

    def _get_root(self):
        return self._root

    root = property(_get_root, doc=textwrap.dedent("""\
        The root of the panel's tree.

        """))

    def _get_parent(self):
        return self._parent

    def _set_parent(self, parent):
        for child in self.descendants:
            assert child is not node, "cannot create loops"
        if self.parent is not None:
            self.parent.remove(self)
        parent.add(self)

    parent = property(_get_parent, _set_parent, doc=textwrap.dedent("""\
        The parent of the panel.

        """))

    def _get_children(self):
        return tuple(self._children)

    children = property(_get_children, doc=textwrap.dedent("""\
        The children of the panel.

        """))

    def _get_visible(self):
        return self._visible

    def _set_visible(self, visible):
        self._visible = visible

    visible = property(_get_visible, _set_visible, doc=textwrap.dedent("""\
        The visibility flag to shortcut drawing and signal handling.

        """))

    ## Child Handling
    #################

    @property
    def descendants(self):
        """Iterator over this and descendant nodes.

        """
        yield self
        for child in list(self._children):
            for node in child.descendants:
                yield node

    def add(self, child, **kwds):
        """Add a node as a child of this one.

        :Parameters:
            `child` : Panel
                The panel to add as a child.

        """
        # Ensure we have an instance of Panel.
        if isinstance(child, type) and issubclass(child, Panel):
            child = child()
        child._load_keyword_attributes(**kwds)

        # Make sure the child is valid.
        assert child._parent is None, "node cannot have multiple parents"

        # Incorporate child into the tree.
        child._parent = self
        for node in child.descendants:
            node._root = self._root
        self._children.append(child)

        # Clear all cached data for the subtree.
        self._clear_cached_attributes()

        # Return the new child object.
        return child

    def remove(self, child):
        """Remove a child node from this one.

        :Parameters:
            `child` : Panel
                The child panel to remove.

        """
        self._children.remove(child)
        self._clear_cached_attributes()

    ## Dimension Handling
    #####################

    _symbol_factors = { "right"  :  1.0,  "top"    :  1.0,
                        "center" :  0.5,  "middle" :  0.5,
                        "left"   :  0.0,  "bottom" :  0.0, }

    _sym = lambda s,x: s._symbol_factors.get(x, x)
    _mgn = lambda s,i,x: s.margin[i] - sum(s.margin[i::2]) * s._sym(x)
    _pad = lambda s,i,x: s.padding[i] - sum(s.padding[i::2]) * s._sym(x)

    # The 'maximum view' is simply the largest the view can be. It provides
    # a default for the content size as well as providing upper bounds for
    # the computed view size.

    @attribute.Cached
    def max_view_size(self):
        """The maximum size of the visible panel content.

        """
        # Parent dimensions.
        lwidth, lheight = self.layout_size

        # Margin & padding dimensions.
        hpad = lheight * (sum(self.margin[0::2]) + sum(self.padding[0::2]))
        vpad = lheight * (sum(self.margin[1::2]) + sum(self.padding[1::2]))

        # "Inner" dimensions.
        iwidth = max(0.0, lwidth - hpad)
        iheight = max(0.0, lheight - vpad)

        # Aspect scaling.
        if self.aspect is not None:
            aspectx, aspecty = self.aspect
            if aspectx * iheight < aspecty * iwidth:
                iwidth = iheight * aspectx / aspecty
            elif aspectx * iheight >= aspecty * iwidth:
                iheight = iwidth * aspecty / aspectx

        # 0) Sensible default.
        width, height = iwidth, iheight

        if self.width is not None:
            if self.height is not None:
                # 1) Have width and height.
                width = min(iwidth, lwidth * self.width - hpad)
                height = min(iheight, lheight * self.height - vpad)
            elif self.aspect is not None:
                # 2) Have width and aspect, no height.
                width = min(iwidth, lwidth * self.width - hpad)
                height = width * aspecty / aspectx
            else:
                # 3) Have width sans height and aspect.
                width = min(iwidth, lwidth * self.width - hpad)
                height = iheight

        elif self.height is not None:
            if self.aspect is not None:
                # 4) Have height and aspect, no width.
                height = min(iheight, lheight * self.height - vpad)
                width = height * aspectx / aspecty
            else:
                # 5) Have height sans width and aspect.
                width = iwidth
                height = min(iheight, lheight * self.height - vpad)

        return width, height

    # The 'content' is a subregion of the parent's content and is defined by
    # every (renderable) subclass. It is the exact size of the renderable
    # content, the offset is determined automatically.

    @attribute.Cached
    def content_size(self):
        """The size of the panel content.

        """
        return self.compute_content_size() or self.max_view_size

    @attribute.Cached
    def parent_size(self):
        """The size of the parent panel's content.

        """
        if self.parent is not None:
            return self.parent.content_size
        return (0.0, 0.0)

    @attribute.Cached
    def layout_size(self):
        """The panel's layout dimensions used in relative measurements.

        """
        if isinstance(self.parent, Layout):
            return self.parent.layout_size
        return self.parent_size

    @attribute.Cached
    def content_offset(self):
        """The static (cached) offset from the parent's content corner.

        """
        # Symbol resolvers.
        sym, mgn, pad = self._sym, self._mgn, self._pad

        # Base dimensions.
        cwidth, cheight = self.content_size
        pcwidth, pcheight = self.parent_size

        # Alignment dimensions.
        if self.align_width is not None:
            pcwidth = self.align_width
        if self.align_height is not None:
            pcheight = self.align_height

        # Child alignment displacement.
        offset_x = pcwidth * sym(self.halign)
        offset_y = pcheight * sym(self.valign)

        # Child margin displacement
        offset_x += pcheight * (mgn(0, self.halign) + pad(0, self.halign))
        offset_y += pcheight * (mgn(1, self.valign) + pad(1, self.valign))

        # Child anchor point
        offset_x -= cwidth * sym(self.anchor_x)
        offset_y -= cheight * sym(self.anchor_y)

        return offset_x, offset_y

    @attribute.Cached
    def cumulative_offset(self):
        """The cumulative static offset from the window's content corner.

        """
        if self.parent is not None:
            ox, oy = self.parent.cumulative_offset
            px, py = self.content_offset
            return ox + px, oy + py
        return self.content_offset

    @property
    def dynamic_offset(self):
        """The dynamic (uncached) offset from the parent's content corner.

        """
        return self.xoffset + self.align_xoffset + self.scroll_xoffset, \
               self.yoffset + self.align_yoffset + self.scroll_yoffset

    # The 'view' is a subregion of the parent's view and a subregion of the
    # window. It is the size of the rendered content and as such is never any
    # larger than the content.

    @attribute.Cached
    def view_size(self):
        """The size of the visible panel content.

        """
        cwidth, cheight = self.content_size
        mvwidth, mvheight = self.max_view_size
        return min(cwidth, mvwidth), min(cheight, mvheight)

    # The 'frame' is a region enclosing the view defined by the padding. It is
    # used to render backgrounds and outlines and for scissoring.

    @attribute.Cached
    def frame_size(self):
        """The size of the panel frame.

        """
        lwidth, lheight = self.layout_size
        vwidth, vheight = self.view_size
        hpad = lheight * sum(self.padding[0::2])
        vpad = lheight * sum(self.padding[1::2])
        return vwidth + hpad, vheight + vpad

    @attribute.Cached
    def frame_offset(self):
        """The offset to the frame's corner.

        """
        lwidth, lheight = self.layout_size
        return lheight * -self.padding[0], \
               lheight * -self.padding[1]

    ## Signal Handling
    ##################

    def convert_coordinates(self, x, y):
        """Convert the coordinates from parent relative coordinates.

        :Parameters:
            `x`, `y` : float
                The coordinates to convert.

        """
        static_x, static_y = self.content_offset
        dynamic_x, dynamic_y = self.dynamic_offset
        return x - (static_x + dynamic_x), y - (static_y + dynamic_y)

    def contains_point(self, x, y):
        """Determine whether a point is in the panel.

        :Parameters:
            `x`, `y` : float
                The coordinates to test.

        """
        width, height = self.content_size
        return 0.0 <= x < width and 0.0 <= y < height

    def set_handler(self, name, func, *args):
        """Create a signal handler for the given signal.

        Additional arguments are passed to the function.

        :Parameters:
            `name` : str
                The name of the signal
            `func` : function
                The function to call.

        """
        setattr(self, name, lambda *a: func(*args) or True)
        return self

    def process_signal(self, name, *args):
        """Recurse a signal down the panel tree.

        Additional arguments are passed to the signal handler.

        :Parameters:
            `name` : str
                The name of the signal.

        """
        if self._visible:
            for child in reversed(list(self.children)):
                result = child.process_signal(name, *args)
                if result is not None:
                    return result
            if hasattr(self, name):
                if METAPOD_DEBUG:
                    self.report_signal(name, *args)
                return getattr(self, name)(*args)

    def process_mouse_signal(self, name, x, y, *args):
        """Recurse a mouse signal down the panel tree.

        Additional arguments are passed to the signal handler.

        :Parameters:
            `name` : str
                The name of the signal.
            `x`, `y` : float
                The coordinates of the signal.

        """
        if self._visible:
            x, y = self.convert_coordinates(x, y)
            for child in reversed(list(self.children)):
                result = child.process_mouse_signal(name, x, y, *args)
                if result is not None:
                    return result
            if hasattr(self, name):
                if METAPOD_DEBUG:
                    self.report_signal(name, x, y, *args)
                return getattr(self, name)(x, y, *args)

    def process_contained_signal(self, name, x, y, *args):
        """Recurse a mouse signal through panels which contain it.

        Additional arguments are passed to the signal handler.

        :Parameters:
            `name` : str
                The name of the signal.
            `x`, `y` : float
                The coordinates of the signal.
        """
        if self._visible:
            x, y = self.convert_coordinates(x, y)
            for child in reversed(list(self.children)):
                result = child.process_contained_signal(name, x, y, *args)
                if result is not None:
                    return result
            if hasattr(self, name) and self.contains_point(x, y):
                if METAPOD_DEBUG:
                    self.report_signal(name, x, y, *args)
                return getattr(self, name)(x, y, *args)

    def report_signal(self, name, *args):
        """Print the given signal and arguments to standard output.

        :Parameters:
            `name` : str
                The name of the signal.

        """
        formats = { float: "%.1f", long: "%.1f", int: "%.1f",
                    unicode: "%r", str: "%r" }
        subs = [formats.get(type(arg), "%r") for arg in args]
        print "%r -- %s(%s)" % (self, name, ", ".join(subs)) % args

    ## Rendering
    ############

    @property
    def scissor_offset(self):
        """The dynamic offset of the panel's scissor box.

        """
        if self.parent is not None:
            parent_x, parent_y = self.parent.scissor_offset
            dynamic_x, dynamic_y = self.dynamic_offset
            return parent_x + dynamic_x, parent_y + dynamic_y
        return self.dynamic_offset

    @attribute.Cached
    def scissor_shape(self):
        """The shape of the scissor rectangle.

        """
        ox, oy = self.cumulative_offset
        fx, fy = self.frame_offset
        fw, fh = self.frame_size
        scx = int(math.ceil(ox+fx))
        scy = int(math.ceil(oy+fy))
        scw = int(math.floor(ox+fx+fw)) - scx
        sch = int(math.floor(oy+fy+fh)) - scy
        while self.parent is not None:
            if self.parent.scissor:
                pscx, pscy, pscw, psch = self.parent.scissor_shape
                scw = min(scx + scw, pscx + pscw)
                sch = min(scy + sch, pscy + psch)
                scx = max(scx, pscx)
                scy = max(scy, pscy)
                scw = scw - scx
                sch = sch - scy
            else:
                self = self.parent
        return scx, scy, scw, sch

    @attribute.Cached
    def frame_vlist(self):
        """The vertex list used to render the frame texture.

        """
        # Frame dimensions.
        fw, fh = self.frame_size
        t = self.frame_width * self.root.content_size[1] / 2

        # Cell processing order.
        frame_cells = [(0, 0), (1, 0), (2, 0), (2, 1),
                       (2, 2), (1, 2), (0, 2), (0, 1)]
        cell_offsets = [(0, 0), (1, 0), (1, 1), (0, 1)]

        # Texture dimensions.
        if self.frame_texture is not None:
            tx, ty = self.frame_texture.tex_coords[:2]
            tw = self.frame_texture.tex_coords[6] - tx
            th = self.frame_texture.tex_coords[7] - ty
        elif self.frame_color is not None:
            tx = ty = tw = th = 0.0

        # Utility mappings.
        txmap = {0: tx,   1: tx+tw/2,   2: tx+tw/2,   3: tx+tw}
        tymap = {0: ty,   1: ty+th/2,   2: ty+th/2,   3: ty+th}
        vxmap = {0: -t,   1:       t,   2:    fw-t,   3:  fw+t}
        vymap = {0: -t,   1:       t,   2:    fh-t,   3:  fh+t}

        # Create data arrays.
        cdata, tdata, vdata = [255] * 128, [], []
        for cx, cy in frame_cells:
            for ox, oy in cell_offsets:
                # Extend data arrays.
                tdata.extend([txmap[cx+ox], tymap[cy+oy]])
                vdata.extend([vxmap[cx+ox], vymap[cy+oy]])
        if self.frame_color is not None:
            cdata = [int(255 * x) for x in self.frame_color] * 32


        # Create and return vertex list.
        if self.frame_texture is not None:
            return pyglet.graphics.vertex_list(32, ('c4B', cdata),
                                                   ('t2f', tdata),
                                                   ('v2f', vdata),)
        elif self.frame_color is not None:
            return pyglet.graphics.vertex_list(32, ('c4B', cdata),
                                                   ('v2f', vdata),)

    @attribute.Cached
    def background_vlist(self):
        """The vertex list used to render the background texture.

        """
        fw, fh = self.frame_size
        cdata, tdata, vdata = [255] * 16, [], []

        # Texture dimensions.
        if self.background_texture is not None:
            tx, ty = self.background_texture.tex_coords[:2]
            tw = self.background_texture.tex_coords[6] - tx
            th = self.background_texture.tex_coords[7] - ty
        elif self.background_color is not None:
            tx = ty = tw = th = 0.0

        # Color data.
        if self.background_color is not None:
            cdata = [int(255 * x) for x in self.background_color] * 4

        tdata = [tx, ty, tx + tw, ty, tx + tw, ty + th, tx, ty + th]
        vdata = [0, 0, fw, 0, fw, fh, 0, fh]

        # Create and return vertex list.
        if self.background_texture is not None:
            return pyglet.graphics.vertex_list(4, ('c4B', cdata),
                                                  ('t2f', tdata),
                                                  ('v2f', vdata),)
        else:
            return pyglet.graphics.vertex_list(4, ('c4B', cdata),
                                                  ('v2f', vdata),)

    def draw_frame(self):
        """Render the panel frame.

        """
        texture = self.frame_texture
        if texture is not None:
            glPushAttrib(GL_TEXTURE_BIT)
            glEnable(texture.target)
            glBindTexture(texture.target, texture.id)
        self.frame_vlist.draw(GL_QUADS)
        if texture is not None:
            glPopAttrib()

    def draw_background(self):
        """Render the panel background.

        """
        texture = self.background_texture
        if texture is not None:
            glPushAttrib(GL_TEXTURE_BIT)
            glEnable(texture.target)
            glBindTexture(texture.target, texture.id)
        self.background_vlist.draw(GL_QUADS)
        if texture is not None:
            glPopAttrib()

    def draw(self):
        """Render the panel and child panels.

        """
        # Shortcut if disabled.
        if not self._visible:
            return

        # Offset to content corner.
        glPushMatrix()
        ox, oy = self.content_offset
        dx, dy = self.dynamic_offset
        glTranslatef(ox + dx, oy + dy, 0.0)

        # Flag to track scissor.
        _have_set_scissor = False

        # Clean computations.
        width, height = self.content_size

        # Render the background.
        if (self.background_texture is not None) or \
           (self.background_color is not None):
            glPushMatrix()
            fx, fy = self.frame_offset
            glTranslatef(fx, fy, 0.0)
            self.draw_background()
            glPopMatrix()

        # Set scissor state.
        if self.scissor:
            _have_set_scissor = True
            glPushAttrib(GL_SCISSOR_BIT)
            scx, scy, scw, sch = self.scissor_shape
            ox, oy = self.scissor_offset
            glScissor(scx + ox, scy + oy, scw, sch)
            glEnable(GL_SCISSOR_TEST)

        # Render the content.
        self.draw_content()

        # Render the children.
        for child in self.children:
            child.draw()

        # Unset scissor state.
        if _have_set_scissor:
            glPopAttrib()

        # Render the frame.
        if (self.frame_texture is not None) or \
           (self.frame_color is not None):
            glPushMatrix()
            fx, fy = self.frame_offset
            glTranslatef(fx, fy, 0.0)
            self.draw_frame()
            glPopMatrix()

        # Render debug rectangles.
        if METAPOD_DEBUG:
            if Panel._debug_focus_panel is self:
                cw, ch = self.content_size
                fx, fy = self.frame_offset
                fw, fh = self.frame_size
                _recttest(0, 0, cw, ch, (1, 0, 0, 1))
                _recttest(fx, fy, fw, fh, (0, 1, 0, 1))

        glPopMatrix()

    ## Debug
    ########

    if METAPOD_DEBUG:
        _debug_focus_panel = None
        def debug_focus(self, x, y):
            Panel._debug_focus_panel = self
            return True


###############################################################################
## LAYOUT PANEL IMPLEMENTATIONS                                              ##
###############################################################################

class Layout(Panel):
    pass
