"""Utility classes and functions.

"""

from __future__ import division

import inspect

import pyglet
from pyglet.gl import *

__all__ = [
    "Vector",
    "get_cls_name"
    "graphicsgroup"
]

class Vector(tuple):
    """Simple 2d floating-point vector implementation.

    """

    def __str__(self):
        """Return a concise string representation.

        """
        return "%s.%s((%.2f, %.2f))" % (inspect.getmodule(self).__name__,
                type(self).__name__, self[0], self[1])

    def __repr__(self):
        """Return a precise string representation.

        """
        return "%s.%s((%r, %r))" % (inspect.getmodule(self).__name__,
                type(self).__name__, self[0], self[1])

    def __neg__(self):
        """Return the unary negation of the vector.

        """
        return type(self)((-self[0], -self[1]))

    def __add__(self, other):
        """Return the component-wise sum of the two vectors.

        """
        return type(self)((self[0] + other[0], self[1] + other[1]))

    def __radd__(self, other):
        """Return the component-wise sum of the two vectors.

        """
        return type(self)((other[0] + self[0], other[1] + self[1]))

    def __sub__(self, other):
        """Return the component-wise difference of the two vectors.

        """
        return type(self)((self[0] - other[0], self[1] - other[1]))

    def __rsub__(self, other):
        """Return the component-wise difference of the two vectors.

        """
        return type(self)((other[0] - self[0], other[1] - self[1]))

    def __mul__(self, other):
        """Return the component-wise product of the vector with a scalar.

        """
        return type(self)((self[0] * other, self[1] * other))

    def __rmul__(self, other):
        """Return the component-wise product of the vector with a scalar.

        """
        return type(self)((other * self[0], other * self[1]))

    def __div__(self, other):
        """Return the component-wise division of the vector by a scalar.

        """
        return type(self)((self[0] / other, self[1] / other))

    def __truediv__(self, other):
        """Return the component-wise division of the vector by a scalar.

        """
        return type(self)((self[0] / other, self[1] / other))

    def __floordiv__(self, other):
        """Return the component-wise floored division of the vector by a scalar.

        """
        return type(self)((self[0] // other, self[1] // other))

def get_cls_name(obj):
    return "%s.%s" % (inspect.getmodule(obj).__name__, type(obj).__name__)

class _GraphicsGroup(pyglet.graphics.Group):
    def __init__(self, func, parent=None):
        super(_GraphicsGroup, self).__init__(parent)
        self.func = func
        self.gen = None
    def set_state(self):
        self.gen = self.func()
        for _ in self.gen: break
    def unset_state(self):
        for _ in self.gen: break
def graphicsgroup(func=None, parent=None):
    def _graphicsgroup(func):
        return _GraphicsGroup(func, parent)
    if func is None:
        return _graphicsgroup
    return _graphicsgroup(func)
