from __future__ import division

from collections import defaultdict
from vector import v

from common import *
from constants import *

level_string = """
#####........###.timer!.###........#####
#...#........###........###........#...#
#..............#........#..............#
#....#...#.....##########.....##.......#
.....#...#..................####........
........................................
........................................
............####........####.....###....
..###.....###..............###...###....
............#..............#............
............#..###....###..#............
..#............#........#...............
..#...###......#........#...............
..#.....#......................####.....
..###...#......................#........
........#......................#........
........#......................#........
...............#........#...............
...............#........#..........#....
............#..###....###..#.......#....
..#.........#..............#.......#....
..#..#....###..............###.....#....
..####......####........####.....###....
........................................
.................................###....
.....#...#..............................
#....#...#...##############............#
#............#............#............#
#...#........#............#........#...#
#####........#.score here.#........#####
"""

class ProjectileIndex(object):
    cellsize = 30
    def __init__(self):
        self.clear()
    def clear(self):
        self.buckets = defaultdict(list)
    def key(self, pos):
        cx = int(pos[0] / self.cellsize)
        cy = int(pos[1] / self.cellsize)
        return (cx, cy)
    def add(self, p):
        self.buckets[self.key(p.pos)].append(p)
    def get(self, pos, radius):
        for cx in xrange(int((pos[0] - radius) / self.cellsize), int((pos[0] + radius) / self.cellsize) + 1):
            for cy in xrange(int((pos[1] - radius) / self.cellsize), int((pos[1] + radius) / self.cellsize) + 1):
                for p in self.buckets[(cx, cy)]:
                    yield p
    def colliders(self, ent):
        r = ent.radius + EXTRA_PROJECTILE_RADIUS
        for p in self.get(ent.pos, r):
            if p.pos[0] < ent.pos[0] - r or p.pos[0] > ent.pos[0] + r: continue
            if p.pos[1] < ent.pos[1] - r or p.pos[1] > ent.pos[1] + r: continue
            yield p

class ObstacleIndex(object):
    cellsize = 30
    def __init__(self):
        self.buckets = defaultdict(list)
    def keys(self, p1, p2):
        for xx in xrange(int(p1[0] / self.cellsize), int(p2[0] / self.cellsize) + 1):
            for yy in xrange(int(p1[1] / self.cellsize), int(p2[1] / self.cellsize) + 1):
                yield (xx, yy)
    def add_block(self, block):
        for k in self.keys(v(block.x1, block.y1), v(block.x2, block.y2)):
            self.buckets[k].append(block)
    def get_obstacles(self, pos, radius):
        res = []
        r = v((radius, radius))
        for k in self.keys(pos - r, pos + r):
            res += self.buckets[k]
        return res
        
     

class Level(object):
    xmin = 10
    ymin = 10
    xmax = 630
    ymax = 470

    def __init__(self):
        self.core = Block(280, 210, 360, 270)
        self.blocks = [self.core]

        for yy, line in enumerate(level_string.strip().splitlines()):
            for xx, char in enumerate(line.strip()):
                if char == '#':
                    self.blocks.append(Block(xx*16, yy*16, xx*16 + 16, yy*16 + 16))

        self.index = ObstacleIndex()
        for b in self.blocks:
            self.index.add_block(b)

        self.spawns = [
                        v(480, 490), #      1 o'clock
                        v(650, 320), #      2 o'clock
                        v(650, 160), #      4 o'clock
                        v(480, -10), #      5 o'clock
                        v(160, -10), #      7 o'clock
                        v(-10, 160), #      8 o'clock
                        v(-10, 320), #      10 o'clock  
                        v(160, 490), #      11 o'clock

                      ]
                       
        self.calculate_paths()

    def contains(self, pos, radius):
        return pos[0] - radius >= self.xmin and pos[0] + radius <= self.xmax and pos[1] - radius >= self.ymin and pos[1] + radius <= self.ymax

    def solid(self, pos):
        for b in self.index.get_obstacles(pos, 0):
            if b is self.core and hasattr(b, 'hp') and b.hp <= 0: continue
            if pos[0] <= b.x1: continue
            if pos[0] >= b.x2: continue
            if pos[1] <= b.y1: continue
            if pos[1] >= b.y2: continue
            return True
        else:
            if pos[0] < self.xmin or pos[0] > self.xmax: return True
            if pos[1] < self.ymin or pos[1] > self.ymax: return True
            return False

    def push(self, pos, radius):
        for b in self.index.get_obstacles(pos, radius):
            pos = b.push(pos, radius)
        return pos

    def push_in(self, pos, radius):
        if self.contains(pos, radius):
            return pos
        return v((max(min(pos[0], self.xmax-radius), self.xmin+radius), max(min(pos[1], self.ymax-radius), self.ymin+radius)))

    attack_points = None
    def attack_point(self, pos):
        rx = min(max(pos[0], self.core.x1), self.core.x2)
        ry = min(max(pos[1], self.core.y1), self.core.y2)
        return v((rx, ry))

    def in_attack_zone(self, pos, a_range=ATTACK_ZONE_SIZE):
        if pos[0] < self.core.x1 - a_range or pos[0] > self.core.x2 + a_range: return False
        if pos[1] < self.core.y1 - a_range or pos[1] > self.core.y2 + a_range: return False
        return True

    def calculate_paths(self):
        self.paths = defaultdict(lambda: 0)
        queue = []
        cx1, cy1 = pos_to_cpos(v((self.core.x1, self.core.y1)))
        cx2, cy2 = pos_to_cpos(v((self.core.x2, self.core.y2)))
        for cx in xrange(cx1, cx2 + 1):
            for cy in xrange(cy1, cy2 + 1):
                cpos = v((cx, cy))
                self.paths[cpos] = 1000
                for n in neighbours(cpos):
                    queue.append((n, 999))
        while queue:
            cpos, h = queue.pop(0)
            if cpos in self.paths: continue
            if self.solid(cpos_to_pos(cpos)): continue
            self.paths[cpos] = h
            for n in neighbours(cpos):
                queue.append((n, h-1))
                
    def get_path_dir(self, pos):
        cc = pos_to_cpos(pos)
        best_dest = max(neighbours(cc), key=self.paths.get)
        return (cpos_to_pos(best_dest) - pos).safe_normalised()

class Block(object):
    def __init__(self, x1, y1, x2, y2):
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2
        self.xm = (x1 + x2) / 2
        self.ym = (y1 + y2) / 2
    
    def push(self, pos, radius):
        if pos[0] <= self.x1-radius or pos[0] >= self.x2+radius: return pos
        if pos[1] <= self.y1-radius or pos[1] >= self.y2+radius: return pos
        opts = [v((pos[0], self.y1-radius)), v((pos[0], self.y2+radius)),                
                v((self.x1-radius, pos[1])), v((self.x2+radius, pos[1]))]
        return min(opts, key=lambda p: (p - pos).length2)
