import pyglet
from pyglet.window import key
from pyglet.event import EVENT_HANDLED
from pyglet.event import EVENT_UNHANDLED

import mode
import interface
import texts
import gamemode
import vector
import random
import data

from pyglet.gl import *

import config
import gamedata
from common import *
from constants import *


class Thing(object):
    def __init__(self, h, img, pos, heading):
        self.h = h
        self.img = img
        self.tex = data.load_texture(self.img)
        self.pos = pos
        self.old_pos = pos
        self.target = pos
        self.heading = heading
        self.friction = 0.8
        self.thrust = 2
        self.age = 0.01
        self.vlist = None
        self.radius = 5
        self.scale = 1
        self.update_vlist()
    def update_vlist(self):
        if self.vlist is None:
            self.create_vlist()
        from math import sin, cos, radians
        x, y = self.pos
        h = radians(self.heading)
        r = self.radius
        rc = r * self.scale * cos(h)
        rs = r * self.scale * sin(h)
        self.vlist.vertices[:] = [
                x-rc+rs, y-rs-rc,
                x+rc+rs, y+rs-rc,
                x+rc-rs, y+rs+rc,
                x-rc-rs, y-rs+rc,
        ]
    def create_vlist(self):
        tc = self.tex.tex_coords
        vdata = [0] * 8
        cdata = [255, 255, 255, 127] * 4
        tdata = [tc[i] for i in (0, 1, 3, 4, 6, 7, 9, 10)]
        group = pyglet.graphics.TextureGroup(self.tex)
        self.vlist = self.h.batch.add(4, GL_QUADS, group, ('v2f', vdata), ('t2f', tdata), ('c4B', cdata))
    def tick(self):
        o = self.pos
        self.pos += (self.pos - self.old_pos) * self.friction
        self.pos += (self.target - self.pos).safe_scaled_to(self.thrust)
        self.old_pos = o
        self.heading = (self.pos - self.old_pos).angle
        self.age += 1
        if random.random() < 0.1 or (self.target - self.pos).length < 2:
            width = 640; height = 480
            x = max(0, min(self.pos.x + random.randrange(-300, 300), width))
            y = max(0, min(self.pos.y + random.randrange(-300, 300), height))
            self.target = vector.v((x, y))
        self.update_vlist()


class TextMenuItem(interface.content.Text):

    callback = interface.attribute.Attribute(None)


class TextMenu(interface.layout.Vertical):

    selected = None

    selected_color = interface.attribute.Attribute((1.0, 1.0, 0.0, 1.0))
    unselected_color = interface.attribute.Attribute((1.0, 1.0, 1.0, 1.0))
    item_options = interface.attribute.Attribute({})

    def add_item(self, text, callback):
        opts = dict(self.item_options)
        opts["color"] = self.unselected_color
        return self.add(TextMenuItem, text=text, callback=callback, **opts)

    def add_spacer(self, height):
        self.add(interface.content.Color,height=height,width=.1,color=[0,0,0,0])

    def change_selection(self, selection):
        if self.selected is not None:
            child = self.children[self.selected]
            self.selected = None
            if isinstance(child, TextMenuItem):
                child.color = self.unselected_color
        if selection is not None:
            child = self.children[selection]
            self.selected = selection
            if isinstance(child, TextMenuItem):
                child.color = self.selected_color

    def move_selection(self, adj):
        top = len(self.children) - 1
        if self.selected is None:
            if adj < 0:
                self.change_selection(top)
            else:
                self.change_selection(0)
        else:
            idx = max(0, min(top, self.selected + adj))
            self.change_selection(idx)
        if not isinstance(self.children[self.selected], TextMenuItem):
            self.move_selection(adj)

    def activate_selection(self):
        child = self.children[self.selected]
        if isinstance(child.callback, tuple):
            function, args = child.callback
            function(*args)
        elif child.callback is not None:
            child.callback()

    def key_press(self, symbol, modifiers):
        if symbol in (pyglet.window.key.UP, pyglet.window.key.W):
            self.move_selection(-1)
        elif symbol in (pyglet.window.key.DOWN, pyglet.window.key.S):
            self.move_selection(+1)
        elif symbol in mode.FIRE:
            self.activate_selection()
        else:
            return
        return True


class MenuMode(mode.Mode):
    name = "menu"

    def __init__(self):
        super(MenuMode, self).__init__()

        self.root = interface.content.Root(
                aspect = (4, 3),
                scissor = True,
                background_color = (0.1, 0.1, 0.1, 1.0),
        )

        self.batch = pyglet.graphics.Batch()
        self.world = self.root.add(gamemode.World)
        self.world.draw_world = self.batch.draw

        self.title = self.root.add(interface.content.Text,
                text = "Loopback",
                font_size = 0.08,
                font_name = FONT_NAME,
                color = (1, 1, 1, 1),
                valign = "top",
        )

        self.menu = self.root.add(TextMenu,
                item_options = MENU_FONT_OPTS,
        )

        self.menu.add_item("Start Game", self.do_show_start)
        self.menu.add_spacer(0.05)
        self.opt_fs = self.menu.add_item("Protect the core!", self.do_toggle_fullscreen)
        self.opt_sm = self.menu.add_item("Love the core!", self.do_toggle_sound)
        self.menu.add_item("Help & Credits", self.do_help)
        self.menu.add_item("Quit Game", self.do_quit_game)
        self.menu.change_selection(0)

        self.dont_toggle_fullscreen()
        self.dont_toggle_sound()

        self.difficulty = self.root.add(TextMenu,
                item_options = MENU_FONT_OPTS,
                visible = False,
        )

        self.difficulty.add_item("Easy", (self.do_start_game, [3, "Easy"]))
        self.difficulty.add_item("Medium", (self.do_start_game, [6, "Medium"]))
        self.difficulty.add_item("Hard", (self.do_start_game, [10, "Hard"]))
        self.difficulty.add_item("Very Hard", (self.do_start_game, [15, "Very Hard"]))
        self.difficulty.add_item("Extreme", (self.do_start_game, [25, "Extreme"]))
        self.difficulty.add_item("Impossible", (self.do_start_game, [40, "Impossible"]))
        self.difficulty.change_selection(1)

        self.info = self.root.add(interface.content.Text,
                font_name = FONT_NAME,
                font_size = 0.025,
                text = texts.INFO_TEXT,
                visible = False,
                width = 0.8,
        )

        self.highscore = self.root.add(interface.content.Text,
                font_name = FONT_NAME,
                font_size = 0.04,
                text = "High Score: %06d" % config.highscore,
                valign = "bottom",
        )

        self.things = []

        self.handler = interface.event.Handler(self.root)

    def do_show_start(self):
        self.menu.visible = False
        self.difficulty.visible = True

    def do_toggle_fullscreen(self):
        new_value = not config.fullscreen
        self.control.window.set_fullscreen(new_value)
        config.save_option("fullscreen", new_value)
        self.window.set_mouse_visible(not config.fullscreen)
        self.dont_toggle_fullscreen()

    def dont_toggle_fullscreen(self):
        self.opt_fs.text = "Mode: fullscreen" if config.fullscreen else "Mode: windowed"

    def do_toggle_sound(self):
        new_value = not config.sound
        if new_value:
            self.control.music.random_play(MUSICS)
        else:
            self.control.music.stop()
        config.save_option("sound", new_value)
        self.dont_toggle_sound()
        
    def dont_toggle_sound(self):
        self.opt_sm.text = "Sound & music: on" if config.sound else "Sound & music: off"

    def do_help(self):
        self.menu.visible = False
        self.info.visible = True

    def do_back(self):
        self.menu.visible = True
        self.info.visible = False
        self.difficulty.visible = False

    def do_quit_game(self):
        pyglet.app.exit()

    def do_start_game(self, difficulty, diff):
        gd = gamedata.GameData(difficulty, diff)
        self.control.switch_handler("game", gd)

    def on_key_press(self, symbol, modifiers):
        if symbol == key.ESCAPE:
            if self.menu.visible:
                self.do_quit_game()
            else:
                self.do_back()
        elif symbol == key.ENTER:
            if self.info.visible:
                self.do_back()
        else:
            return EVENT_UNHANDLED
        return EVENT_HANDLED

    def connect(self, control):
        super(MenuMode, self).connect(control)
        self.window.push_handlers(self.handler)
        self.window.push_handlers(self.keys)
        self.root.window = self.window

    def disconnect(self):
        self.window.remove_handlers(self.keys)
        self.window.pop_handlers()
        super(MenuMode, self).disconnect()

    def tick(self):
        for t in list(self.things):
            t.tick()
        if len(self.things) < 20:
            width, height = 640, 480
            pos = vector.v((random.random() * width, random.random() * height))
            heading = random.random() * 360
            enemies = ["blue.png", "cyan.png", "green.png", "lilac.png", "orange.png", "pink.png", "purple.png", "red.png", "yellow.png"]
            img = random.choice(enemies)
            self.things.append(Thing(self, img, pos, heading))

    def on_draw(self):
        self.window.clear()
        self.root.draw()
